#ifndef __NGX_WASM_H_INCLUDED__
#define __NGX_WASM_H_INCLUDED__

/*
 * Copyright (C) 2024 Web Server LLC
 */

#include <ngx_config.h>
#include <ngx_core.h>


#define ngx_wasm_create_runtime   ngx_wasm_actions.create_runtime
#define ngx_wasm_delete_runtime   ngx_wasm_actions.delete_runtime
#define ngx_wasm_create_module    ngx_wasm_actions.create_module
#define ngx_wasm_delete_module    ngx_wasm_actions.delete_module
#define ngx_wasm_create_instance  ngx_wasm_actions.create_instance
#define ngx_wasm_delete_instance  ngx_wasm_actions.delete_instance
#define ngx_wasm_enter            ngx_wasm_actions.enter
#define ngx_wasm_translate        ngx_wasm_actions.translate


#define NGX_WASM_MODULE           0x4D634157  /* "MSAW" */
#define NGX_WASM_CONF             0x02000000

#define ngx_wasm_get_conf(conf_ctx, module)                                   \
    ngx_get_conf(conf_ctx, ngx_wasm_module)                                   \
            ? (*(ngx_get_conf(conf_ctx, ngx_wasm_module))) [module.ctx_index] \
            : NULL


typedef struct ngx_wasm_instance_s ngx_wasm_instance_t;


/* configuration of the 'wasm_modules' block */
typedef struct {
    ngx_uint_t              use;         /* configured runtime */
    u_char                 *name;        /* name of configured runtime */
    ngx_array_t             modules;     /* of ngx_wasm_module_t    */

#if defined(NGX_HTTP_WASM)
    /*
     * indexes of modules marked with subsys=http go here;
     * no real dependency on HTTP here except the fact that http
     * subsystem will use this module list
     */
    ngx_array_t             http_modules; /* of ngx_uint_t */
#endif

} ngx_wasm_conf_t;


/* loadable WASM module */
typedef struct {
    ngx_str_t               ns;          /* namespace for exported functions */
    ngx_str_t               file;        /* file name is in config */
    ngx_str_t               src;         /* file contents */
    ngx_str_t               init_func;   /* module conf init function name */
    ngx_uint_t              index;       /* in array of all loaded modules */
    ngx_wasm_conf_t        *wcf;
    void                   *data;        /* loaded module */
    void                   *ctx;         /* loaded module runtime config */
    ngx_wasm_instance_t    *reactor;     /* persistent module instance */
    ngx_array_t            *acl;         /* allowed APIs*/
    ngx_array_t            *preopen;     /* allowed host directories */
    ngx_uint_t              is_reactor;  /* unsigned is_reactor:1; */
} ngx_wasm_module_t;


#include <ngx_wasm_host.h>

/* wasm call arguments */
typedef enum {
    NGX_WT_I32,
    NGX_WT_I64,
    NGX_WT_F32,
    NGX_WT_F64,
    /* TODO: refs */
} ngx_wasm_val_kind_t;


typedef struct {
    union {
        int32_t            i32;
        int64_t            i64;
        float              f32;
        double             f64;
    } of;
    ngx_wasm_val_kind_t    kind;
} ngx_wasm_val_t;


/* WASM function's arguments and environment */
typedef struct {
    ngx_wasm_host_t        *host;   /* pointer to host */
    ngx_str_t               name;   /* function to call: wasi:bar */
    uint32_t                nargs;  /* arguments count */
    ngx_wasm_val_t         *args;   /* arguments */
    ngx_wasm_val_t          rc;     /* return value */
} ngx_wasm_call_t;


/* WASM module's instance */
struct  ngx_wasm_instance_s {
    ngx_pool_t             *pool;          /* pool to use for instance alloc */
    ngx_log_t              *log;           /* instance log */
    ngx_wasm_module_t      *mod;           /* instance origin */
    void                   *data;          /* instance (native to runtime) */
    void                   *ctx;           /* context for instance's owner */
};


/* separate module that provides host functions environment */

/* methods/properties of WASM runtimes */
typedef struct {
    ngx_int_t (*create_runtime)(ngx_cycle_t *cycle);
    void      (*delete_runtime)(ngx_cycle_t *cycle);

    ngx_int_t (*create_module)(ngx_cycle_t *cycle, ngx_wasm_module_t *wm);
    void      (*delete_module)(ngx_wasm_module_t *wm);


    ngx_wasm_instance_t  *(*create_instance)(ngx_wasm_module_t *wm,
                                             ngx_pool_t *pool, ngx_log_t *log);
    void                  (*delete_instance)(ngx_wasm_instance_t *wi);

    ngx_int_t             (*enter)(ngx_wasm_instance_t *wi,
                                   ngx_wasm_call_t *call);

    void     *(*translate)(ngx_wasm_instance_t *wi, uint32_t addr, size_t len);

    ngx_uint_t  validate_on_start;
} ngx_wasm_actions_t;


/* module implementing WASM runtime */
typedef struct {
    ngx_str_t              *name;

    void                 *(*create_conf)(ngx_cycle_t *cycle);
    char                 *(*init_conf)(ngx_cycle_t *cycle, void *conf);

    ngx_wasm_actions_t      actions;
} ngx_wasm_core_module_t;


/* obtain module by name */
ngx_wasm_module_t *ngx_wasm_get_module(ngx_conf_t *cf, ngx_str_t *ns);

/* for use with directives */
typedef struct {
    ngx_wasm_module_t      *wm;   /* wasm module reference */
    ngx_str_t              func;  /* wasm function name */
} ngx_wasm_func_conf_t;

typedef struct {
    ngx_str_t               host;
    ngx_str_t               guest;
} ngx_wasm_preopen_t;

char *ngx_conf_set_wasm_ref_slot(ngx_conf_t *cf, ngx_command_t *cmd,
    void *conf);


extern ngx_module_t         ngx_wasm_module;
extern ngx_module_t         ngx_wasm_core_module;
extern ngx_wasm_actions_t   ngx_wasm_actions;


#endif /* __NGX_WASM_H_INCLUDED__ */
